==> frgal-3.14.0-py3-none-any.whl: frgal/utils.py <==
"""Various utilities."""
import asyncio
import random
import string
import warnings
from functools import partial
from threading import Thread
from typing import Any, AsyncIterable, Awaitable, Callable, ClassVar, Iterable, List, Optional, Sequence, TypeVar
from unittest.mock import Mock

import grpc
from aioitertools.itertools import list as alist
from google.protobuf.message import Message
from google.protobuf.text_format import MessageToString

MESSAGE_MAX_LENGTH = 256

T = TypeVar('T')


def make_awaitable(value: T) -> Awaitable[T]:
    """Convert value to an awaitable."""
    async def _wrapper() -> T:
        return value
    return _wrapper()


def get_call_id() -> str:
    """Return random call ID for logging."""
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(4))  # nosec


def format_message(message: Any) -> str:
    """Format a message to string.

    If argument is a gRPC Message, format it nicely.
    If it's another object, use simple str conversion.
    In any case the output has limited length.
    """
    if isinstance(message, Message):
        out = MessageToString(message, as_one_line=True)
    else:
        out = str(message)
    if len(out) > MESSAGE_MAX_LENGTH:
        return out[:MESSAGE_MAX_LENGTH - 3] + '...'
    else:
        return out


class CallableMock:
    """Mock of a gRPC method Callable object."""

    def __init__(self, method: str, call: Callable, *, exhaust_iterables: Optional[bool] = None):
        self._method = method.encode('utf-8')
        self.call = call
        self.exhaust_iterables = exhaust_iterables

    def __call__(self, request: Any, *args: Any, **kwargs: Any) -> Any:
        """Mock gRPC method call."""
        if self.exhaust_iterables:
            request = self._exhaust(request)
        elif (self.exhaust_iterables is None and not isinstance(request, Sequence)
              and isinstance(request, (Iterable, AsyncIterable))):
            warnings.warn(
                'Default value of `exhaust_iterables` will change to True in next major release.'
                'If you want to maintain the current behaviour, set it to False explicitely.',
                DeprecationWarning,
            )

        return self.call(request, *args, **kwargs)

    def future(self, *args: Any, **kwargs: Any) -> Mock:
        """Mock asynchronous gRPC method call."""
        future_mock = Mock(name='future', spec=('result', ))
        future_mock.result.return_value = self.call(*args, **kwargs)
        return future_mock

    @staticmethod
    def _exhaust(value: Any) -> Any:
        """Exhaust any iterables."""
        results: List[Any] = []
        exceptions: List[Exception] = []

        def _inner(value: AsyncIterable, results: List[Any], exceptions: List[Exception]) -> Any:
            try:
                results.append(tuple(asyncio.run(alist(value))))
            except Exception as err:
                exceptions.append(err)

        if isinstance(value, Sequence):
            # Strings, lists and tuples should be left untouched
            return value
        elif isinstance(value, Iterable):
            return tuple(value)
        elif isinstance(value, AsyncIterable):
            # Run async iterable in separate thead – this works both
            # with and without running async loop in the main thread
            thread = Thread(target=_inner, args=(value, results, exceptions))
            thread.start()
            thread.join()
            if exceptions:
                raise exceptions[0]
            else:
                return results[0]
        else:
            return value


class AsyncMock(Mock):
    """Async mock variant."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Async mock call."""
        result = super().__call__(*args, **kwargs)
        if isinstance(result, (AsyncIterable, Awaitable)):
            return result
        else:
            warnings.warn(
                'Automagical transformation to awaitables is deprecated '
                'and will be removed in the next major release.',
                DeprecationWarning,
            )
            return make_awaitable(result)


class MockingChannel(grpc.Channel):
    """A channel which binds Stub calls to TestClient internal mock."""

    def __init__(self, client: 'TestClientMixin', exhaust_iterables: Optional[bool] = None):
        self.client = client
        self.exhaust_iterables = exhaust_iterables

    def subscribe(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""

    def unsubscribe(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""

    def unary_unary(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def unary_stream(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def stream_unary(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def stream_stream(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def close(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""


class TestClientMixin:
    """Mixin to wrap gRPC client for tests.

    Attributes:
        mock: A mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    exhaust_iterables: ClassVar[Optional[bool]] = None

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.mock = Mock(spec=('__call__', ))

    def connect(self) -> None:
        """Create a mocking channel which redirects all gRPC method calls to the mock attribute."""
        self.channel = MockingChannel(self, exhaust_iterables=self.exhaust_iterables)


class AsyncTestClientMixin(TestClientMixin):
    """Mixin to wrap async gRPC client for tests.

    Attributes:
        mock: An async mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.mock = AsyncMock(spec=('__call__', ))
        self._channel = MockingChannel(self, exhaust_iterables=self.exhaust_iterables)  # type: Any
        self._channel._loop = asyncio.new_event_loop()

    def __del__(self) -> None:
        self._channel._loop.close()

==> frgal-3.14.2-py3-none-any.whl: frgal/utils.py <==
"""Various utilities."""
import asyncio
import random
import string
import warnings
from functools import partial
from threading import Thread
from typing import Any, AsyncIterable, Awaitable, Callable, ClassVar, Iterable, List, Optional, Sequence, TypeVar
from unittest.mock import Mock

import grpc
from aioitertools.builtins import list as alist
from google.protobuf.message import Message
from google.protobuf.text_format import MessageToString

MESSAGE_MAX_LENGTH = 256

T = TypeVar('T')


def make_awaitable(value: T) -> Awaitable[T]:
    """Convert value to an awaitable."""
    async def _wrapper() -> T:
        return value
    return _wrapper()


def get_call_id() -> str:
    """Return random call ID for logging."""
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(4))  # nosec


def format_message(message: Any) -> str:
    """Format a message to string.

    If argument is a gRPC Message, format it nicely.
    If it's another object, use simple str conversion.
    In any case the output has limited length.
    """
    if isinstance(message, Message):
        out = MessageToString(message, as_one_line=True)
    else:
        out = str(message)
    if len(out) > MESSAGE_MAX_LENGTH:
        return out[:MESSAGE_MAX_LENGTH - 3] + '...'
    else:
        return out


class CallableMock:
    """Mock of a gRPC method Callable object."""

    def __init__(self, method: str, call: Callable, *, exhaust_iterables: Optional[bool] = None):
        self._method = method.encode('utf-8')
        self.call = call
        self.exhaust_iterables = exhaust_iterables

    def __call__(self, request: Any, *args: Any, **kwargs: Any) -> Any:
        """Mock gRPC method call."""
        if self.exhaust_iterables:
            request = self._exhaust(request)
        elif (self.exhaust_iterables is None and not isinstance(request, Sequence)
              and isinstance(request, (Iterable, AsyncIterable))):
            warnings.warn(
                'Default value of `exhaust_iterables` will change to True in next major release.'
                'If you want to maintain the current behaviour, set it to False explicitely.',
                DeprecationWarning,
            )

        return self.call(request, *args, **kwargs)

    def future(self, *args: Any, **kwargs: Any) -> Mock:
        """Mock asynchronous gRPC method call."""
        future_mock = Mock(name='future', spec=('result', ))
        future_mock.result.return_value = self.call(*args, **kwargs)
        return future_mock

    @staticmethod
    def _exhaust(value: Any) -> Any:
        """Exhaust any iterables."""
        results: List[Any] = []
        exceptions: List[Exception] = []

        def _inner(value: AsyncIterable, results: List[Any], exceptions: List[Exception]) -> Any:
            try:
                results.append(tuple(asyncio.run(alist(value))))
            except Exception as err:
                exceptions.append(err)

        if isinstance(value, Sequence):
            # Strings, lists and tuples should be left untouched
            return value
        elif isinstance(value, Iterable):
            return tuple(value)
        elif isinstance(value, AsyncIterable):
            # Run async iterable in separate thead – this works both
            # with and without running async loop in the main thread
            thread = Thread(target=_inner, args=(value, results, exceptions))
            thread.start()
            thread.join()
            if exceptions:
                raise exceptions[0]
            else:
                return results[0]
        else:
            return value


class AsyncMock(Mock):
    """Async mock variant."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Async mock call."""
        result = super().__call__(*args, **kwargs)
        if isinstance(result, (AsyncIterable, Awaitable)):
            return result
        else:
            warnings.warn(
                'Automagical transformation to awaitables is deprecated '
                'and will be removed in the next major release.',
                DeprecationWarning,
            )
            return make_awaitable(result)


class MockingChannel(grpc.Channel):
    """A channel which binds Stub calls to TestClient internal mock."""

    def __init__(self, client: 'TestClientMixin', exhaust_iterables: Optional[bool] = None):
        self.client = client
        self.exhaust_iterables = exhaust_iterables

    def subscribe(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""

    def unsubscribe(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""

    def unary_unary(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def unary_stream(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def stream_unary(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def stream_stream(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def close(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""


class TestClientMixin:
    """Mixin to wrap gRPC client for tests.

    Attributes:
        mock: A mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    exhaust_iterables: ClassVar[Optional[bool]] = None

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.mock = Mock(spec=('__call__', ))

    def connect(self) -> None:
        """Create a mocking channel which redirects all gRPC method calls to the mock attribute."""
        self.channel = MockingChannel(self, exhaust_iterables=self.exhaust_iterables)


class AsyncTestClientMixin(TestClientMixin):
    """Mixin to wrap async gRPC client for tests.

    Attributes:
        mock: An async mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.mock = AsyncMock(spec=('__call__', ))
        self._channel = MockingChannel(self, exhaust_iterables=self.exhaust_iterables)  # type: Any
        self._channel._loop = asyncio.new_event_loop()

    def __del__(self) -> None:
        self._channel._loop.close()


async def _consume_async_gen(gen: Callable[..., AsyncIterable], *args: Any, **kwargs: Any) -> list:
    return await alist(gen(*args, **kwargs))
